/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

/*  ----- from tinyproxy for jftpgw
 *
 * Logs the various messages which tinyproxy produces to either a log file or
 * the syslog daemon. Not much to it...
 *
 * Copyright (C) 1998  Steven Young
 * Copyright (C) 1999  Robert James Kaes (rjkaes@flarenet.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * log.c - For the manipulation of log files.
 */

#include <syslog.h>
#include "ftp_main.h"

void ftp_log_cmd_to_target(struct cmdlogtarget* ltarget, struct proxyinfo* pi) {
	struct clientinfo* ci = pi->ci;

	char* commandpattern;
	size_t commandpatternsize;
	time_t nowtime;
	char* ws;

	commandpatternsize = strlen(ci->lastcmd.cmd) + 3;
	commandpattern = (char*) malloc(commandpatternsize);
	enough_mem(commandpattern);

	if ((ws = strpbrk(ci->lastcmd.cmd, " \t")) == NULL) {
		/* Command consisting of a single word */
		snprintf(commandpattern, commandpatternsize, " %s ", ci->lastcmd.cmd);
	} else {
		commandpattern[0] = ' ';
		strncpy(&commandpattern[1], ci->lastcmd.cmd, ws - ci->lastcmd.cmd);
		commandpattern[ws-ci->lastcmd.cmd+1] = ' ';
		commandpattern[ws-ci->lastcmd.cmd+2] = '\0';
	}
	toupstr(commandpattern);

	if (incommandpattern(ltarget->specs, commandpattern)) {
		char* replaced;
		const char* line_pattern;
		/* found, log it */
		nowtime = time(NULL);

		if (strcmp(ltarget->style, "commonlog") == 0) {
			/* %D is [18/Feb/2001:11:42:46 CET] */
			line_pattern = "%a - %l %D \"%r\" %s %b";
			/* instead of the - there should be the the
			 * authenticated id if we require a login/password in
			 * order to use the proxy. */
		} else if (strcmp(ltarget->style, "xferlog") == 0) {
			/* %t is Wed Feb 14 01:41:28 2001 */
			line_pattern =
				"%t %T %H %b \"%f\" %y _ %w %o %u %e 0 * %c";
		} else {
			line_pattern = ltarget->style;
		}
		replaced = log_replace_line(line_pattern, pi);
		if (ltarget->syslog.used) {
			if (ltarget->syslog.prefix) {
				syslog(ltarget->syslog.priority, "%s: %s\n",
							ltarget->syslog.prefix,
							replaced);
			} else {
				syslog(ltarget->syslog.priority, "%s\n",
							replaced);
			}
		} else {
			fprintf(ltarget->logf, "%s\n", replaced);
			fflush(ltarget->logf);
		}
		free(replaced);
	}
	free(commandpattern);
}


char* ftp_log_replace_char(const char pattern, struct proxyinfo* pi) {
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;
	union log_replace_un replace_val;
	int replace_type;
	time_t nowtime;
	struct timeval tv;
	char* str;

	switch (pattern) {
		case 'c': /* complete */
			replace_val.replace_char
				= ci->data_connection.complete ? 'c' : 'i';
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'D': /* common log date */
			str = (char*) malloc(LOG_STRFTIME_LENGTH);
			enough_mem(str);
			nowtime = time(NULL);
			/* this is also defined for commonlog log format */
			strftime(str, LOG_STRFTIME_LENGTH,
				"[%d/%b/%Y:%H:%M:%S %Z]",
				localtime(&nowtime));
			replace_val.replace_str = str;
			replace_type = LOG_REPLACE_STRING_MALLOCED;
			break;
		case 't': /* like Wed Feb 14 01:41:28 2004 */
			str = (char*) malloc(LOG_STRFTIME_LENGTH);
			enough_mem(str);
			nowtime = time(NULL);
			/* this is also defined for xferlog log format */
			strftime(str, LOG_STRFTIME_LENGTH,
				"%a %b %d %H:%M:%S %Y",
				localtime(&nowtime));
			replace_val.replace_str = str;
			replace_type = LOG_REPLACE_STRING_MALLOCED;
			break;
		case 'T': /* Time taken to transmit/receive file, in seconds */
			replace_val.replace_uint
				= ci->data_connection.transfer_duration;
			replace_type = LOG_REPLACE_UINT;
			break;
		case 'b': /* Bytes sent for request */
			replace_val.replace_luint
				= ci->data_connection.transferred;
			replace_type = LOG_REPLACE_LUINT;
			break;
		case 'R': /* throughput rate in kbyte/s */
			if (ci->data_connection.transfer_duration) {
				replace_val.replace_float =
					(ci->data_connection.transferred / 1024) /
						ci->data_connection.transfer_duration;
				replace_type = LOG_REPLACE_FLOAT;
			} else {
				replace_val.replace_char = '-';
				replace_type = LOG_REPLACE_CHAR;
			}
			break;
		case 'f': /* Filename stored or retrieved, absolute path (if
			   * given as absolute path) */
			if (ci->lastcmd.filename) {
				replace_val.replace_str = ci->lastcmd.filename;
				replace_type = LOG_REPLACE_STRING;
			} else {
				replace_val.replace_char = '-';
				replace_type = LOG_REPLACE_CHAR;
			}
			break;
		case 'F':
			/* Filename stored or retrieved, as the client sees
			 * it base_name is just a pointer within lcs->filename
			 * */
			if (ci->lastcmd.filename) {
				replace_val.replace_str
					= base_name(ci->lastcmd.filename);
				replace_type = LOG_REPLACE_STRING;
			} else {
				replace_val.replace_char = '-';
				replace_type = LOG_REPLACE_CHAR;
			}
			break;
		case 'm': /* Command (method) name received from client,
			     e.g., RETR */
			replace_val.replace_str
				= ci->lastcmd.method;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'r': /* full commandline */
			replace_val.replace_str
				= ci->lastcmd.cmd;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'P': /* pid */
			replace_val.replace_uint
				= getpid();
			replace_type = LOG_REPLACE_UINT;
			break;
		case 's': /*  Numeric FTP response code (status) */
			replace_val.replace_int
				= ci->lastcmd.respcode;
			replace_type = LOG_REPLACE_INT;
			break;
		case 'y': /* tYpe */
			replace_val.replace_char
				= ci->data_connection.asciiconversion.type;
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'w':  /* direction - show 'i' for UNSPEC as well */
			replace_val.replace_char
				= ci->data_connection.direction
					== TRANSFER_DIRECTION_UPSTREAM ? 'o':'i';
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'o':  /* anonymous? */
			/* if we have an anonymous login, anon_user_pass
			 * contains the password of the user. Then we return
			 * 'a'. If we have a real authenticated user, we have
			 * "UNKNOWN" in anon_user_pass. Then we return 'r' */
			if (!ci->user || !ci->anon_user_pass) {
				replace_val.replace_char = '-';
			} else if (strcmp(ci->anon_user_pass, "UNKNOWN") == 0) {
				replace_val.replace_char = 'r';
			} else {
				replace_val.replace_char = 'a';
			}
			replace_type = LOG_REPLACE_CHAR;
			break;
		case 'e': /* sErvice */
			replace_val.replace_str = ci->service;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'n': /* aNon-user */
			replace_val.replace_str = ci->anon_user_pass;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'u': /* User-id */
			/* UNKNOWN if anonymous user is logged in
			 * proxy username if authenticated user is logged in */
			if (!ci->user || !ci->anon_user_pass) {
				replace_val.replace_str = "-";
			} else if (strcmp(ci->anon_user_pass, "UNKNOWN") == 0) {
				replace_val.replace_str = ci->user_login;
			} else {
				replace_val.replace_str = ci->anon_user_pass;
			}
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'H': /* Server host name */
			replace_val.replace_str = si->connection.server_name;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'A': /* Server host IP */
			replace_val.replace_str
				= conv_ip2char(si->connection.server_ip);
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'h': /* Client host name */
			replace_val.replace_str = si->connection.client_name;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'a': /* Client host IP */
			replace_val.replace_str
				= conv_ip2char(si->connection.client_ip);
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'I': /* Server interface address */
			replace_val.replace_str
				= conv_ip2char(si->connection.addr_to_server);
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'i': /* Client interface address */
			replace_val.replace_str
				= conv_ip2char(si->connection.addr_to_client);
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'l': /* Proxy user name (login) */
			replace_val.replace_str = ci->user_login;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'L': /* Effective server user name, i.e. user name that
			     the proxy sends to the server "USER x", so we log
			     x here */
			replace_val.replace_str = ci->user;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'C': /* Forwarded server user name, i.e. the user name the
			     forwarding rules constructed. If you have a
			     forward rule like:

			     forward *@ftp.fu-berlin.de
			     
			     and you log in as anonymous, the forwarding
			     mechanism will create a login:
			     "anonymous@ftp.fu-berlin.de:21"
			*/
			replace_val.replace_str = ci->forward.login;
			replace_type = LOG_REPLACE_STRING;
			break;
		case 'U': /* unix time in seconds since the epoch (1970-01-01) */
			replace_val.replace_luint = (unsigned long int) nowtime;
			replace_type = LOG_REPLACE_LUINT;
			break;
		case 'M': /* unix time in seconds since the epoch (1970-01-01)
			     microseconds included after a dot */
			gettimeofday(&tv, NULL);
			str = (char*) malloc(LOG_STRFTIME_LENGTH);
			snprintf(str, LOG_STRFTIME_LENGTH,
					"%lu.%lu", tv.tv_sec, tv.tv_usec);
			replace_val.replace_str = str;
			replace_type = LOG_REPLACE_STRING_MALLOCED;
			break;
		case '%': /* percent sign */
			replace_val.replace_char = '%';
			replace_type = LOG_REPLACE_CHAR;
			break;
		default:
			return (char*) 0;
	}
	return log_replace(replace_type, replace_val);
}

