/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "ftp_main.h"
#include "ftp_cmds.h"

/* XXX before the transfer, tell the server to use binary mode - we do this
 * before EVERY transfer now. But it's not the best solution... */

int std_quit(const char* cmd, const char** args,
		struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct ftp_message msg;
	char* msg_no_codes;

	/* Are we already connected to a server? */
	if (ci->login.stage < LOGIN_ST_CONNECTED) {
		/* Generate an own goodbye message if we are
		 * not yet connected */
		message_buffer_add(msgbuf, cs, 221, "Goodbye...\r\n");
		return CMD_QUIT;
	}
	say(ss, "QUIT\r\n");
	msg = ftp_readall(ss);
	if (msg.error) {
		ftp_message_say_error(msg, msgbuf, cs, __FILE__, __LINE__);
		/* we ignore the error */
		message_buffer_add(msgbuf, cs, 221, "Goodbye...\r\n");
		return CMD_QUIT;
	}
	/* got valid msg */
	msg_no_codes = strip_respcodes(msg.msg);
	message_buffer_add_nomalloc(msgbuf, cs,
			221, msg_no_codes);
	ftp_message_free(&msg);

	/* we quit even if the server shows a message "500 I won't let you go"
	 * for example  :-)
	 * That's why we also send the 221 message and not the one of the
	 * server */
	return CMD_QUIT;
}


int std_pasv(const char* cmd, const char** args,
		struct proxyinfo* pi, struct messagebuffer* msgbuf) {

	int ret = pasvclient(pi, msgbuf);
	/* XXX add more detailed error evaluation */
	if (ret) {
		return CMD_ERROR;
	}

	return CMD_HANDLED;
}


int std_port(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct sockaddr_in sin;

	ASSUME(args != NULL);

	if (ci->data_connection.portcmd) {
		free_and_zero_char(&ci->data_connection.portcmd);
	}
	if (portcommandcheck(args[0], &sin, pi, msgbuf) < 0) {
		/* an error was already reported */
		return CMD_ERROR;
	}
	ci->data_connection.transfermode.clientmode = ACTIVE;
	ci->data_connection.portcmd = strdup(args[0]);
	enough_mem(ci->data_connection.portcmd);

	message_buffer_add(msgbuf, cs,
				200, "PORT command successful.\r\n");
	return CMD_HANDLED;
}

int ftp_transfer(struct proxyinfo* pi, struct messagebuffer* msgbuf,
				int *source_fd, int *target_fd,
				int direction,
				transfermode_func_t t_in_f,
				transfermode_func_t t_out_f,
				transfermode_deinit_func_t deinit_first_f,
				transfermode_deinit_func_t deinit_second_f) {

	struct clientinfo* ci = pi->ci;
	struct transfer_t t;
	struct ftp_message msg = ftp_message_empty();
	size_t transferred;
	unsigned int transfer_duration;
	time_t transfer_start_time;
	int ret;
	int cs;
	int ss;

	cs = pi->si->connection.clientsocket;
	ss = pi->si->connection.serversocket;

	ci->data_connection.direction = direction;

	/* actually the filter may pump up or shrink the size of the
	 * transferred bytes... But we focus on the client side because this is
	 * the side where we do logging for. */
	transferred = 0;
	t.bufsize = 0;
	t.buffer = (char*) 0;
	transfer_start_time = time(NULL);

	do {
		ret = t_in_f(*source_fd, &t, pi);
		if (ret < 0) {
			break;
		}
		/* upstream means we get it from the client and push it to
		 * the server. So here we count the bytes the client sends
		 * */
		if (direction == TRANSFER_DIRECTION_UPSTREAM) {
			transferred += t.bufsize;
		}
		/* direction is defined by falling back to the filter
		 * constants, so there is no problem with different values */
		ret = filters_apply(direction, &t, pi);
		if (ret < 0) {
			break;
		}
		ret = t_out_f(*target_fd, &t, pi);
		if (ret < 0) {
			break;
		}
		/* downstream means we get it from the server and send it to
		 * the client. So here we count the bytes the client receives
		 * */
		if (direction == TRANSFER_DIRECTION_DOWNSTREAM) {
			transferred += t.bufsize;
		}
	} while (t.status == TRANSFER_STATUS_NORMAL);

	transfer_duration = time(NULL) - transfer_start_time;
	pi->ci->data_connection.transferred = transferred;
	pi->ci->data_connection.transfer_duration = transfer_duration;

	/* in case of an error the buffer gets already freed by the
	 * error-throwing function and in case of a finished transfer there is
	 * nothing left to free as well because buffer will be 0 */

	ret = deinit_first_f(pi, msgbuf, &msg);
	if (ret < 0) {
		/* XXX */
		return -1;
	}

	ret = deinit_second_f(pi, msgbuf, &msg);
	if (ret < 0) {
		/* XXX */
		return -1;
	}

	if (ret < 0) {
		message_buffer_add(msgbuf, cs,
				500, "Error\r\n"); /* XXX */
		return CMD_ERROR;
	} else {
	/*	Transfer ok	*/	
	}

	return 0;
}

/* ---- RETR ---- */
int ftp_retr(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf,
			const char* ctm_name) {

	struct clientinfo* ci = pi->ci;
	const struct transfermode_t* stm, *ctm;
	int least_id;
	int ret;
	struct ftp_message msg = ftp_message_empty();

	/* this function is also called by ftp_list, so we do not necessarily
	 * have a filename */
	ASSUME(args != NULL);
	if (args[0]) {
		ci->lastcmd.filename = strdup(args[0]);
	} else {
		ci->lastcmd.filename = strdup("-");
	}
	enough_mem(ci->lastcmd.filename);

	/* set up the client with the requested transfer mode */
	ctm = transfermode_get_by_name(TRANSFERMODE_SEND_CLIENT, ctm_name);

	/* we try to find a suitable transfermode for the connection to the
	 * server */
	least_id = -1;
	do {
		stm = transfermode_get(TRANSFERMODE_RECV_SERVER, least_id);
		if (stm) { least_id = stm->id; }
		ret = stm->transfermode_init_func(cmd, args,
							pi, msgbuf, &msg);
		if (ret == TRANSFERMODE_FAIL ||
		    ret == TRANSFERMODE_FATAL ||
		    ret == TRANSFERMODE_OK) {
			/* break the loop */
				break;
		}
		if (ret == TRANSFERMODE_NOTSUPPORTED) {
			ftp_message_free(&msg);
			/* loop again */
		}
	} while (stm != (struct transfermode_t*) 0);

	if (stm == (struct transfermode_t*) 0) {
		jlog(2, "No suitable transfermode found in std_retr");
		return CMD_ERROR;
	}

	if (ret == TRANSFERMODE_FATAL) {
		jlog(2, "Transfermode requests shutdown of session");
		return CMD_ABORT;
	}

	if (ret == TRANSFERMODE_FAIL) {
		/* tell the client that it didn't work */
		ret = ctm->transfermode_sorry_client_func(cmd, args, pi,
								msgbuf, &msg);
		return CMD_ERROR;
	}

	/* ok, we have ret == TRANSFERMODE_OK here */

	ret = ctm->transfermode_init_func(cmd, args, pi, msgbuf, &msg);
	if (ret < 0) {
		jlog(6, "Could not initialize transfermode to client");
		return CMD_ERROR;
	}
	if (msg.code) {
		ftp_message_free(&msg);
	}

	ret = ftp_transfer(pi, msgbuf,
			&(ci->data_connection.dataserversock),  /* fd in  */
			&(ci->data_connection.dataclientsock),  /* fd out */
			TRANSFER_DIRECTION_DOWNSTREAM,
			stm->transfermode_func,              /*     in    */
			ctm->transfermode_func,              /*     out   */
			stm->transfermode_deinit_func,      /* deinit 1st */
			ctm->transfermode_deinit_func);     /* deinit 2nd */

	return ret;
}

int std_retr(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	return ftp_retr(cmd, args, pi, msgbuf, "std");
}

/* ---- STOR ---- */
int ftp_stor(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf,
			const char* ctm_name) {

	struct clientinfo* ci = pi->ci;
	const struct transfermode_t* stm, *ctm;
	int least_id;
	int ret;
	struct ftp_message msg = ftp_message_empty();

	/* set up the client with the requested transfer mode */
	ctm = transfermode_get_by_name(TRANSFERMODE_RECV_CLIENT, ctm_name);

	/* we try to find a suitable transfermode for the connection to the
	 * server */
	least_id = -1;
	do {
		stm = transfermode_get(TRANSFERMODE_SEND_SERVER, least_id);
		if (stm) { least_id = stm->id; }
		ret = stm->transfermode_init_func(cmd, args,
							pi, msgbuf, &msg);
		if (ret == TRANSFERMODE_FAIL ||
		    ret == TRANSFERMODE_FATAL ||
		    ret == TRANSFERMODE_OK) {
			/* break the loop */
				break;
		}
		if (ret == TRANSFERMODE_NOTSUPPORTED) {
			ftp_message_free(&msg);
			/* loop again */
		}
	} while (stm != (struct transfermode_t*) 0);

	if (stm == (struct transfermode_t*) 0) {
		jlog(2, "No suitable transfermode found in std_retr");
		return CMD_ERROR;
	}

	if (ret == TRANSFERMODE_FATAL) {
		jlog(2, "Transfermode requests shutdown of session");
		return CMD_ABORT;
	}

	if (ret == TRANSFERMODE_FAIL) {
		/* tell the client that it didn't work - when the quota is full
		 * or the directory write protected for example */
		ret = ctm->transfermode_sorry_client_func(cmd, args, pi,
								msgbuf, &msg);
		return CMD_ERROR;
	}

	/* ok, we have ret == TRANSFERMODE_OK here */

	ret = ctm->transfermode_init_func(cmd, args, pi, msgbuf, &msg);
	if (ret < 0) {
		jlog(6, "Could not initialize transfermode to client");
		return CMD_ERROR;
	}
	if (msg.code) {
		ftp_message_free(&msg);
	}

	ret = ftp_transfer(pi, msgbuf,
			&(ci->data_connection.dataclientsock),  /* fd in  */
			&(ci->data_connection.dataserversock),  /* fd out */
			TRANSFER_DIRECTION_UPSTREAM,
			ctm->transfermode_func,              /*     in    */
			stm->transfermode_func,              /*     out   */
			stm->transfermode_deinit_func,      /* deinit 1st */
			ctm->transfermode_deinit_func);     /* deinit 2nd */

	return ret;
}

int std_stor(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	return ftp_stor(cmd, args, pi, msgbuf, "std");
}

int std_stou(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	/* fall back to ftp_stor */
	return ftp_stor(cmd, args, pi, msgbuf, "std");
}

/* ---- LIST ---- */

int ftp_list(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf,
			const char* ctm_name) {
	int ret;
	pi->ci->data_connection.asciiconversion.serverlisting = 1;
	ret = ftp_retr(cmd, args, pi, msgbuf, ctm_name);
	pi->ci->data_connection.asciiconversion.serverlisting = 0;
	return ret;
}

int std_list(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	return ftp_list(cmd, args, pi, msgbuf, "std");
}

/* a simple function that determines the transfer mode (ascii or image) */
int std_type(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	int cs = pi->si->connection.clientsocket;
	char type;

	if (strcasecmp(args[0], "a") == 0) {
		type = 'A';
		pi->ci->data_connection.asciiconversion.type = 'a';
	} else {
		type = 'I';
		pi->ci->data_connection.asciiconversion.type = 'i';
	}
	/* activate the filters */
	if (type == 'A') {
		filter_activate("asciiconversion", FILTER_TYPE_DOWNSTREAM);
		filter_activate("asciiconversion", FILTER_TYPE_UPSTREAM);
	} else {
		filter_deactivate("asciiconversion", FILTER_TYPE_DOWNSTREAM);
		filter_deactivate("asciiconversion", FILTER_TYPE_UPSTREAM);
	}
	message_buffer_addf(msgbuf, cs,
			200, "Type set to %c.\r\n", type);
	return CMD_HANDLED;
}

/* Commands that we just forward on ... */

int std_forward_cmd(const char* cmd, const char** args, struct proxyinfo* pi,
			struct messagebuffer *msgbuf, int expected_code) {

	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct ftp_message msg;
	char* msg_no_codes;
	int i;

	/* At least for SITE we have more than one argument */
	ASSUME(args != NULL);
	say(ss, cmd);
	for (i = 0; args[i]; i++) {
		sayf(ss, " %s", args[i]);
	}
	say(ss, "\r\n");
	msg = ftp_readall(ss);
	if (msg.error) {
		ftp_message_say_error(msg, msgbuf, cs, __FILE__, __LINE__);
		return CMD_ABORT;
	}
	msg_no_codes = strip_respcodes(msg.msg);
	if (expected_code != -1 && msg.code != expected_code) {
		if (msg.code < 500 || msg.code > 600) {
			msg.code = 500;
		}
		term_before_last_newline(msg_no_codes);
		message_buffer_addf(msgbuf, cs,
			msg.code, "%s\r\n", msg_no_codes);
		ftp_message_free(&msg);
		return CMD_ERROR;
	}
	message_buffer_add_nomalloc(msgbuf, cs,
				msg.code, msg_no_codes);
	ftp_message_free(&msg);
	return CMD_HANDLED;
}

int std_config_forward_cmd(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	/* we don't know about the return code... */
	return std_forward_cmd(cmd, args, pi, msgbuf, -1);
}

int std_pwd(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 257);
}

int std_cwd(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 250);
}

int std_cdup(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 250);
}

int std_size(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 213);
}

int std_mkd(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 257);
}

int std_rmd(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 250);
}

int std_rnfr(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 350);
}

int std_rnto(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 250);
}

int std_noop(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 200);
}

int std_mdtm(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 213);
}

int std_dele(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 250);
}

int std_site(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	/* 214 is not the only valid return value, though it seems to be used
	 * most often. We leave that one open here... */
	return std_forward_cmd(cmd, args, pi, msgbuf, -1);
}

int std_rest(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	return std_forward_cmd(cmd, args, pi, msgbuf, 350);
}


int std_already_loggedin(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {
	int cs = pi->si->connection.clientsocket;
	message_buffer_add(msgbuf, cs,
			503, "You are already logged in!\r\n");
	return CMD_HANDLED;
}


int std_capa(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer *msgbuf) {

	int cs = pi->si->connection.clientsocket;
	char* capabilities = capa_get_as_char();
	message_buffer_add_nomalloc(msgbuf, cs,
				200, capabilities);

	return CMD_HANDLED;
}

/* --------- transfermode functions ------------ */

int transfermode_std_in_from_client_init(const char* cmd,
					 const char** args,
					 struct proxyinfo* pi,
					 struct messagebuffer* msgbuf,
					 struct ftp_message* server_msg) {

	return transfermode_std_out_to_client_init(cmd, args,
						pi, msgbuf, server_msg);
}

int transfermode_std_in_from_server_init(const char* cmd,
					 const char** args,
					 struct proxyinfo* pi,
					 struct messagebuffer* msgbuf,
					 struct ftp_message* server_msg) {

	int servermode = pi->ci->data_connection.transfermode.servermode;
	int clientmode = pi->ci->data_connection.transfermode.clientmode;
	int ret;
	int i;
	int ss = pi->si->connection.serversocket;
	struct ftp_message msg;

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);
	/* UNSPEC should not exist here because we replace it with a default
	 * during the login. */
	ASSUME(servermode == PASSIVE
		|| servermode == ACTIVE
		|| servermode == ASCLIENT);

	if (clientmode == UNSPEC) {
		*server_msg = ftp_message_construct(425,
				"Unable to build data connection\r\n");
		return CMD_ERROR;
	}

	/* if the client transfers with SPP, it will set PASSIVE as
	 * transfermode. */
	ASSUME(clientmode == PASSIVE || clientmode == ACTIVE);

	if (servermode == ACTIVE
			|| (servermode == ASCLIENT && clientmode == ACTIVE)) {

		ret = activeserver(pi, msgbuf, server_msg);
		ftp_message_free(server_msg);
		if (ret == TRANSFERMODE_FAIL || ret == TRANSFERMODE_FATAL) {
			return ret;
		}
		ASSUME(ret == TRANSFERMODE_OK);
	}
	if (servermode == PASSIVE
			|| (servermode == ASCLIENT && clientmode == PASSIVE)) {

		ret = pasvserver(pi, msgbuf, server_msg);
		ftp_message_free(server_msg);
		if (ret == TRANSFERMODE_FAIL || ret == TRANSFERMODE_FATAL) {
			return ret;
		}
		ASSUME(ret == TRANSFERMODE_OK);
	}

	/* set the mode to be binary (always) */
	say(ss, "TYPE I\r\n");
	msg = ftp_readall(ss);
	if (msg.error) {
		/* XXX */
		ftp_message_free(&msg);
		return TRANSFERMODE_FATAL;
	}
	if (msg.code != 200) {
		/* XXX */
		ftp_message_free(&msg);
		return TRANSFERMODE_FAIL;
	}
	ftp_message_free(&msg);

	/* At least for SITE we have more than one argument */
	ASSUME(args != NULL);
	say(ss, cmd);
	for (i = 0; args[i]; i++) {
		sayf(ss, " %s", args[i]);
	}
	say(ss, "\r\n");

	*server_msg = ftp_readall(ss);
	if (server_msg->error) {
		return TRANSFERMODE_FATAL;
	}

	if (server_msg->code != 150 && server_msg->code != 125) {
		jlog(4, "Server returned invalid response: %s",
							server_msg->msg);
		return TRANSFERMODE_FAIL;
	}

	if (servermode == ACTIVE
			|| (clientmode == ACTIVE && servermode == ASCLIENT)) {

		ret = transfer_wait_connection(&pi->ci->data_connection.dataserversock, (int*) 0);
		if (ret < 0) {
			close(pi->si->connection.serversocket);
			pi->si->connection.serversocket = -1;
			/* XXX really fatal ? */
			return TRANSFERMODE_FATAL;
		}
		if (config_get_bool("allowforeigndataconnect", 0) == 0) {
			unsigned long int control_peer;
			unsigned long int data_peer;

			control_peer = get_uint_peer_ip(pi->si->connection.serversocket);
			data_peer = get_uint_peer_ip(pi->ci->data_connection.dataserversock);
			if (control_peer != data_peer) {
				struct ip_t c_ip, d_ip;
				char* s_ip_c, *s_ip_d;
				c_ip = ip_from_s_addr(control_peer);
				d_ip = ip_from_s_addr(data_peer);
				s_ip_c = ip_to_mchar(c_ip);
				s_ip_d = ip_to_mchar(d_ip);
				jlog(4, "Error: control connection was "
				  "established to %s but data connection came "
				  "from %s", s_ip_c, s_ip_d);
				jlog(4, "Set the option \'allowforeigndataconnect\' to true in order to change this");
				free(s_ip_c);
				free(s_ip_d);
				close(pi->si->connection.serversocket);
				pi->si->connection.serversocket = -1;
				return TRANSFERMODE_FAIL;
			}
		}
	} else if (servermode == PASSIVE) {
		/* we connect to the server */
	}

	return TRANSFERMODE_OK;
}

int transfermode_std_out_to_client_init(const char* cmd,
					const char** args,
					struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message* server_msg) {

	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	int clientmode = ci->data_connection.transfermode.clientmode;
	int respcode;
	int ret = 0;

	ASSUME(clientmode == PASSIVE || clientmode == ACTIVE);

	ASSUME(server_msg != NULL);
	if (server_msg->code) {
		ftp_message_strip_respcode(server_msg);
	} else {
		/* XXX check it */
		*server_msg = ftp_message_construct(150,
				"Opening data connection for transfer.");
	}

	message_buffer_add(msgbuf, cs,
			server_msg->code, server_msg->msg);
	message_buffer_say_reset(msgbuf, cs, &respcode);

	if (clientmode == PASSIVE) {
		ret = transfer_wait_connection(&pi->ci->data_connection.dataclientsock, (int*) 0);
		if (ret < 0) {
			close(pi->ci->data_connection.dataclientsock);
			pi->ci->data_connection.dataclientsock = -1;
			jlog(5, "Error waiting for connection");
			return TRANSFERMODE_FAIL;
		}
		if (config_get_bool("allowforeigndataconnect", 0) == 0) {
			unsigned long int control_peer;
			unsigned long int data_peer;

			control_peer = get_uint_peer_ip(pi->si->connection.clientsocket);
			data_peer = get_uint_peer_ip(pi->ci->data_connection.dataclientsock);
			if (control_peer != data_peer) {
				struct ip_t c_ip, d_ip;
				char* s_ip_c, *s_ip_d;
				c_ip = ip_from_s_addr(control_peer);
				d_ip = ip_from_s_addr(data_peer);
				s_ip_c = ip_to_mchar(c_ip);
				s_ip_d = ip_to_mchar(d_ip);
				jlog(4, "Error: control connection was "
				  "established to %s but data connection came "
				  "from %s", s_ip_c, s_ip_d);
				jlog(4, "Set the option \'allowforeigndataconnect\' to true in order to change this");
				free(s_ip_c);
				free(s_ip_d);
				close(pi->ci->data_connection.dataclientsock);
				pi->ci->data_connection.dataclientsock = -1;
				message_buffer_add(msgbuf, cs,
					425, "Can't establish connection\r\n");
				return TRANSFERMODE_FAIL;
			}
		}
	}
	if (clientmode == ACTIVE && ci->data_connection.portcmd) {
		ret = activeclient(ci->data_connection.portcmd, pi, msgbuf);
		free_and_zero_char(&ci->data_connection.portcmd);
		if (ret) {
			jlog(2, "Error in activeclient()");
			return TRANSFERMODE_FAIL;
		}
	}
	return TRANSFERMODE_OK;
}

int transfermode_std_out_to_server_init(const char* cmd,
					const char** args,
					struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message* server_msg) {

	return transfermode_std_in_from_server_init(cmd, args,
						pi, msgbuf, server_msg);
}

int transfermode_std_deinit_server(struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message* server_msg) {

	int ss = pi->si->connection.serversocket;

	ASSUME(server_msg != NULL);
	/* we get an empty ftp message that we should fill */
	ASSUME(server_msg->code == 0);

	close(pi->ci->data_connection.dataserversock);
	pi->ci->data_connection.dataserversock = -1;

	*server_msg = ftp_readall(ss);
	if (server_msg->error) {
		return CMD_ABORT;
	}

	pi->ci->data_connection.complete = (server_msg->code == 226);

	ASSUME(server_msg != NULL);
	ASSUME(server_msg->code);

	return TRANSFERMODE_OK;
}

int transfermode_std_deinit_client(struct proxyinfo* pi,
					struct messagebuffer* msgbuf,
					struct ftp_message* server_msg) {

	int cs = pi->si->connection.clientsocket;
	int respcode;

	ASSUME(server_msg != NULL);
	if (server_msg->code) {
		ftp_message_strip_respcode(server_msg);
	} else {
		*server_msg = ftp_message_construct(226, "Transfer ok.");
	}

	message_buffer_add(msgbuf, cs,
				server_msg->code, server_msg->msg);
	message_buffer_say_reset(msgbuf, cs, &respcode);
	pi->ci->lastcmd.respcode = respcode;
	ftp_message_free(server_msg);

	close(pi->ci->data_connection.dataclientsock);
	pi->ci->data_connection.dataclientsock = -1;
	/* will be set at the next PORT or PASV */
	pi->ci->data_connection.transfermode.clientmode = UNSPEC;

	return TRANSFERMODE_OK;
}

int transfermode_std_sorry_client(const char* cmd,
				  const char** args,
				  struct proxyinfo* pi,
				  struct messagebuffer* msgbuf,
				  struct ftp_message* server_msg) {

	int cs = pi->si->connection.clientsocket;

	ASSUME(server_msg != NULL);
	if (server_msg->code) {
		ftp_message_strip_respcode(server_msg);
	} else {
		*server_msg = ftp_message_construct(550,
				"An error occurred during the transfer.");
	}
		
	message_buffer_add(msgbuf, cs,
			server_msg->code, server_msg->msg);
	
	ftp_message_free(server_msg);

	/* XXX do a proper deinit */
	return TRANSFERMODE_OK;
}

int
transfermode_std_in(int fd_in,
		    struct transfer_t* transfer,
		    struct proxyinfo* pi) {
	/* the commandhandler opens up a file descriptor, passes it to us
	 * and we read the data from there. Then we do whatever is necessary
	 * to strip off the communication protocol and write chunks of data
	 * to *transfer. This is then passed through the filters and finally
	 * through the output transfermode. Then we are called again to read
	 * further */

	fd_set readset;
	struct timeval readtime;
	int ret;
	ssize_t count;
	size_t size;

	/* SIG_BLOCK - SIG_UNBLOCK ? Or restart (test for EINTR?) */
	do {
		readtime.tv_sec = config_get_ioption("transfertimeout", 300);
		readtime.tv_usec = 0;

		FD_ZERO(&readset);
		FD_SET(fd_in, &readset);

		ret = select(fd_in + 1, &readset, NULL, NULL, &readtime);
		if (ret < 0 && errno == EINTR) {
			continue;
		}
		if (ret < 0) {
			jlog(2, "select error in transfermode_std_in: %s",
					strerror(errno));
			return -1;
		}
		if (ret == 0) {
			jlog(2, "select timeout in transfermode_std_in: %s",
					strerror(errno));
			return -1;
		}
	} while (ret < 0);

	if (!FD_ISSET(fd_in, &readset)) {
		/* shouldn't happen */
		return -1;
	}

	transfer->buffer = (char*) malloc(TRANSFER_BUFFERSIZE);
	enough_mem(transfer->buffer);
	size = TRANSFER_BUFFERSIZE;
	count = readall(fd_in, transfer->buffer, &size);
	jlog(9, "readall read %d bytes", count);
	if (count < 0) {
		jlog(2, "read() error in transfermode_std_in: %s",
							strerror(errno));
		free(transfer->buffer);
		return -1;
	}
	if (count == 0) {
		free(transfer->buffer);
		transfer->buffer = (char*) 0;
		transfer->bufsize = 0;
		transfer->status = TRANSFER_STATUS_FINISHED;
		return 0;
	}
	if (count < transfer->bufsize) {
		/* we didn't use the whole buffer */
		transfer->buffer = realloc(transfer->buffer, count);
	}
	transfer->bufsize = count;
	transfer->status = TRANSFER_STATUS_NORMAL;
	return 0;
}

int
transfermode_std_out(int fd_out,
		     struct transfer_t* transfer,
		     struct proxyinfo* pi) {

	fd_set writeset;
	struct timeval writetime;
	int ret;
	ssize_t count;

	/* SIG_BLOCK - SIG_UNBLOCK ? Or restart (test for EINTR?) */
	do {
		writetime.tv_sec = config_get_ioption("transfertimeout", 300);
		writetime.tv_usec = 0;

		FD_ZERO(&writeset);
		FD_SET(fd_out, &writeset);

		/* XXX: exceptset ? */
		ret = select(fd_out + 1, NULL, &writeset, NULL, &writetime);
		if (ret < 0 && errno == EINTR) {
			continue;
		}
		if (ret < 0) {
			jlog(2, "select error in transfermode_std_out: %s",
					strerror(errno));
			return -1;
		}
		if (ret == 0) {
			jlog(2, "select timeout in transfermode_std_out: %s",
					strerror(errno));
			return -1;
		}
	} while (ret < 0);

	if (!FD_ISSET(fd_out, &writeset)) {
		/* shouldn't happen */
		return -1;
	}

	count = writeall(fd_out, transfer->buffer, transfer->bufsize);
	free_and_zero_char(&transfer->buffer);
	/* we don't touch the buffer size however, so that the sum can be
	 * calculated for the logging */
	if (count < 0) {
		jlog(2, "writeall() error in transfermode_std_out: %s",
				strerror(errno));
		return -1;
	}
	return 0;
}

