/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "ftp_main.h"

#define TRANSFERMODES_MAX 20

static struct {
	struct transfermode_t transfermode[TRANSFERMODES_MAX];
	int ntransfermodes;
} transfermodes;

int transfermode_init(void) {
	memset(&transfermodes, 0, sizeof(transfermodes));
	transfermodes.ntransfermodes = 0;
	/* register the standard modes */
	/* XXX check return value */
	transfermode_register(50,
			TRANSFERMODE_RECV_SERVER,
			"std",
			transfermode_std_in,
			transfermode_std_in_from_server_init,
			transfermode_std_deinit_server,
			NULL,
			"STD IN from server");

	transfermode_register(50,
			TRANSFERMODE_SEND_CLIENT,
			"std",
			transfermode_std_out,
			transfermode_std_out_to_client_init,
			transfermode_std_deinit_client,
			transfermode_std_sorry_client,
			"STD OUT to client");

	transfermode_register(50,
			TRANSFERMODE_RECV_CLIENT,
			"std",
			transfermode_std_in,
			transfermode_std_in_from_client_init,
			transfermode_std_deinit_client,
			transfermode_std_sorry_client,
			"STD IN from client");

	transfermode_register(50,
			TRANSFERMODE_SEND_SERVER,
			"std",
			transfermode_std_out,
			transfermode_std_out_to_server_init,
			transfermode_std_deinit_server,
			NULL,
			"STD OUT to server");
	return 0;
}

int transfermode_deinit(void) {
	int i;
	for (i = 0; i < transfermodes.ntransfermodes; i++) {
		free(transfermodes.transfermode[i].name);
		free(transfermodes.transfermode[i].description);
	}
	memset(&transfermodes, 0, sizeof(transfermodes));
	return 0;
}

int transfermode_compare(const void* a, const void* b) {
	struct transfermode_t* t_a = (struct transfermode_t*) a;
	struct transfermode_t* t_b = (struct transfermode_t*) b;

	if (t_a->priority == t_b->priority) {
		return 0;
	}
	
	/* we want to sort the array and it should be descending in respect of
	 * priorities. That means that a higher priority is expected to be
	 * more at the beginning of the array. So we have to pretend that the
	 * first element is "lower" than the second one in this case */
	if (t_a->priority > t_b->priority) {
		return -1;
	}

	return 1;
}

int transfermode_register(int priority,
			int type,
			const char* name,
			transfermode_func_t transfermode_func,
			transfermode_init_func_t transfermode_init_func,
			transfermode_deinit_func_t transfermode_deinit_func,
			transfermode_sorry_client_func_t transfermode_sorry_client_func,
			const char* description) {

	struct transfermode_t* tm;
	int i;

	if (transfermodes.ntransfermodes == TRANSFERMODES_MAX) {
		return -1;
	}
	tm = &transfermodes.transfermode[transfermodes.ntransfermodes];

	tm->priority = priority;
	tm->type = type;
	tm->name = strdup(name);
	enough_mem(tm->name);
	tm->transfermode_func = transfermode_func;
	tm->transfermode_init_func = transfermode_init_func;
	tm->transfermode_deinit_func = transfermode_deinit_func;
	tm->transfermode_sorry_client_func = transfermode_sorry_client_func;
	tm->description = strdup(description);
	enough_mem(tm->description);
	transfermodes.ntransfermodes++;

	/* sort the array with descending priorities */
	qsort(&transfermodes.transfermode[0], transfermodes.ntransfermodes,
			sizeof(struct transfermode_t), transfermode_compare);

	/* recalculate ids */
	for (i = 0; i < transfermodes.ntransfermodes; i++) {
		transfermodes.transfermode[i].id = i;
	}
	return 0;
}

const struct transfermode_t* transfermode_get(int type, int least_id) {
	/* return the next transfermode that has an id higher than least_id
	 * because the caller got the one with least_id last time and it didn't
	 * succeed */

	int i;
	for (i = least_id + 1; i < transfermodes.ntransfermodes; i++) {
		if (transfermodes.transfermode[i].type == type) {
			return &transfermodes.transfermode[i];
		}
	}
	return (struct transfermode_t*) 0;
}

const struct transfermode_t* transfermode_get_by_name(int type, const char* name) {
	int i;
	for (i = 0; i < transfermodes.ntransfermodes; i++) {
		if (type == transfermodes.transfermode[i].type
			&& strcmp(transfermodes.transfermode[i].name, name) == 0) {

			return &transfermodes.transfermode[i];
		}
	}
	jlog(2, "Didn't find requested transfermode: %s", name);
	return (struct transfermode_t*) 0;
}

